from pmd_beamphysics.units import mec2, c_light
from pmd_beamphysics.species import charge_state, mass_of

from scipy import interpolate
from scipy.integrate import solve_ivp
from scipy.optimize import brent, brentq
import numpy as np



#----------------------
# Analysis

def accelerating_voltage_and_phase(z, Ez, frequency):
    """
    Computes the accelerating voltage and phase for a v=c positively charged particle in an accelerating cavity field.
    
        Z = \int Ez * e^{-i k z} dz 
        
        where k = omega/c = 2*pi*frequency/c
        
        voltage = abs(Z)
        phase = arg(Z)
    
    Input:
        z  (float array):   z-coordinate array (m)
        Ez (complex array): On-axis complex Ez field array (V/m), oscillating as exp(-i omega t), with omega = 2*pi*frequency
        
    Output:
        voltage, phase in V, radian
    
    """
    c=299792458
    omega = 2*np.pi*frequency
    k = omega/c
    fz =Ez*np.exp(-1j*k*z)
    
    # Integrate
    Z = np.trapz(fz, z)
    
    # Max voltage at phase
    voltage = np.abs(Z)
    phase = np.angle(Z)
    
    return voltage, phase
    
    
    
def track_field_1d(z,
                   Ez,
                   frequency=0,
                   z0=0,
                   pz0=0,
                   t0=0,
                   mc2=mec2,  # electron
                   q0=-1,
                   debug=False,
                   max_step=np.inf,
                  ):
    """
    Tracks a particle in a 1d complex electric field Ez, oscillating as Ez * exp(-i omega t)
    
    Uses scipy.integrate.solve_ivp to track the particle. 
    
    Equations of motion:
    
    $ \frac{dz}{dt} = \frac{pc}{\sqrt{(pc)^2 + m^2 c^4)}} c $ 

    $ \frac{dp}{dt} = q E_z $

    $ E_z = \Re f(z) \exp(-i \omega t) $
    
    
    Parameters
    ----------
    z : array_like      
        positions of the field Ez (m)
    Ez : array_like, complex
        On-axis longitudinal electric field (V/m)
        
    z0 :  float, optional = 0
        initial particle position (m)
    pz0 : float, optional = 0
        initial particle momentum (eV/c)
    t0 : float, optional = 0
        initial particle time (s)
    mc2 : float, optional = mec2
        initial particle mass (eV)
    q0 : float, optional = -1
        initial particle charge (e) (= -1 for electron)
    max_step: float, optional = np.inf
        Maximum timestep for solve_ivp (s)
    debug, bool, optional = False
        If True, Will return the full solution to solve_ivp
    
    """
    
    # Make interpolating function
    field = interpolate.interp1d(z, Ez * q0 * c_light, fill_value='extrapolate')
    zmin = z.min()
    zmax = z.max()
    tmax = 100/frequency
    omega = 2*np.pi*frequency

    # function to integrate
    def fun(t, y):
        z = y[0]
        p = y[1]  
        zdot = p/np.hypot(p, mc2) * c_light
        pdot =  np.real(field(z)*np.exp(-1j*omega*t))
        return np.array([zdot, pdot])
        
    # Events (stopping conditions)
    def went_backwards(t, y):
        return y[0]
    went_backwards.terminal = True
    went_backwards.direction = -1

    def went_max(t, y): 
        return y[0] - zmax
    went_max.terminal = True
    went_max.direction = 1

    # Solve
    y0 =  np.array([z0, pz0])
    sol = solve_ivp(fun, (t0, tmax), y0,
                first_step = 1/frequency/100,
                events=[went_backwards, went_max],
               # vectorized=True)   # Make it slower?
                    max_step=max_step)
           #      max_step=1/frequency/20)
    
    if debug:
        return sol
    
    # Final z, p
    zf = sol.y[0][-1]
    pf = sol.y[1][-1]
    
    return zf, pf





def autophase_field(field_mesh, pz0=0, scale=1, species='electron', tol=1e-9, verbose=False, debug=False):
    """
    Finds the maximum accelerating of a FieldMesh by tracking a particle and using Brent's method from scipy.optimize.brent. 
    
    NOTE: Assumes that field_mesh.Ez[0,0,:] is the on-axis Ez field.
    TODO: generalize
    
    Parameters
    ----------
    fieldmesh : FieldMesh object
    
    pz0 : float, optional = 0 
        initial particle momentum in the z direction, in eV/c
        pz = 0 is a particle at rest. 
    scale : float, optional = 1 
        Additional field scale. 
    species : str, optional = 'electron'
        species to track.
    tol : float, optional = 1e-9
        Tolerence for brent: Stop if between iteration change is less than tol.
        
    debug : bool, optional = False
        If true, will return a function that tracks the field at a given phase in deg. 
    verbose : bool, optional = False
        If true, prints information about the v=c voltage and phase for the initial guess, and function call information.
        
    Returns
    -------
    phase : float
        Maximum accelerating phase in deg
    pz1 : float        
        Final particle momentum in the z direction, in eV/c
    
    """
    
    # Get field on-axis
    z = field_mesh.coord_vec('z')
    Ez = field_mesh.Ez[0,0,:]*scale
    frequency = field_mesh.frequency
    zmin = z.min()
    
    # Get mass and charge state
    mc2= mass_of(species)
    q0 = charge_state(species) # -1 for electrons
    
    # Function for use in brent
    def phase_f(phase_deg):
        zf, pf = track_field_1d(z,
                   Ez,
                   frequency=frequency,
                   z0=zmin,
                   pz0=pz0,
                   t0=phase_deg/360/frequency,
                   mc2=mc2,
                   max_step= 1/frequency/10,
                   q0=q0)
        return pf
    
    if debug:
        return phase_f
    
    # Get a quick estimate, to use in the bracket
    voltage0, phase0 = accelerating_voltage_and_phase(z, q0*Ez, frequency)
    phase0_deg = phase0*180/np.pi
    if verbose:
        print(f'v=c voltage: {voltage0} V, phase: {phase0_deg} deg')    
    
    alg_sign=-1
    phase_range = [phase0_deg-90, phase0_deg+90]
    phase1_deg, pz1, iter, funcalls = brent(lambda x: alg_sign*phase_f(x%360), brack=phase_range, maxiter=250, tol=tol, full_output=True)
    if verbose:
        print(f'    iterations: {iter}')
        print(f'    function calls: {funcalls}')
    
    return phase1_deg %360, alg_sign*pz1




def autophase_and_scale_field(field_mesh, voltage, pz0=0, species='electron', debug=False, verbose=False):
    """
    Finds the maximum accelerating of a FieldMesh.
    
    Uses two iterations of phasing, scaling.
    
    Parameters
    ----------
    fieldmesh : FieldMesh object
    
    voltage : float
        Desired on-crest voltage in V
        
    pz0 : float, optional = 0 
        initial particle momentum in the z direction, in eV/c
        pz = 0 is a particle at rest. 

    species : str, optional = 'electron'
        species to track.
        
    debug : bool, optional = False
        If true, will return a function that tracks the field at a given phase and scale.
        
    verbose : bool, optional = False
        If true, prints information about the v=c voltage and phase for the initial guess, and function call information.
        
    Returns
    -------
    phase : float
        Maximum accelerating phase in deg
    scale : float        
        scale factor for the field      
    
    """
    
    z = field_mesh.coord_vec('z')
    Ez = field_mesh.Ez[0,0,:]
    frequency = field_mesh.frequency
    zmin = z.min()
    
    # Get mass and charge
    mc2= mass_of(species)
    q0 = charge_state(species)
    energy0 = np.hypot(pz0, mc2)
    
    # Get and initial estimate
    voltage0, phase0 = accelerating_voltage_and_phase(z, q0*Ez, frequency)
    # convert to deg
    phase0 = phase0*180/np.pi    
    scale0 = voltage/voltage0
    if verbose:
        print(f'v=c voltage: {voltage0} V, phase: {phase0} deg')  
    
        
    def phase_scale_f(phase_deg, scale):
        zf, pf = track_field_1d(z,
                   Ez*scale,
                   frequency=frequency,
                   z0=zmin,
                   pz0=pz0,
                   t0=phase_deg/360/frequency,
                   mc2=mc2,
                   max_step= 1/frequency/10,
                   q0=q0)
        
        delta_energy = np.hypot(pf, mc2) - energy0
        
        return delta_energy
    
    if debug:
        return phase_scale_f
    
    # Phase 1
    brack = [phase0-90, phase0+90]
    phase1 = brent(lambda x: -phase_scale_f(x, scale0), brack=brack, maxiter=250, tol=1e-6, full_output=False) %360
    
    # Scale 1
    s0 = scale0*0.9
    s1 = scale0*1.1
    scale1 = brentq(lambda x: phase_scale_f(phase1, x)/voltage - 1.0, s0, s1, maxiter=20, rtol=1e-6, full_output=False)
    
    if verbose:
        print(f'    Pass 1 delta energy: {phase_scale_f(phase1, scale1)} at phase  {phase1} deg')

    # Phase 2
    brack = [phase1-10, phase1+10]
    phase2 = brent(lambda x: -phase_scale_f(x, scale1), brack=brack, maxiter=250, tol=1e-9, full_output=False) %360    
    
    # Scale 2
    s0 = scale1*0.9
    s1 = scale1*1.1
    scale2 = brentq(lambda x: phase_scale_f(phase2, x)/voltage - 1.0, s0, s1, maxiter=20, rtol=1e-9, full_output=False)    
    
    if verbose:
        print(f'    Pass 2 delta energy: {phase_scale_f(phase2, scale2)} at phase  {phase2} deg')
    
    return phase2, scale2


import numpy as np







def fstr(s):
    """
    Makes a fixed string for h5 files
    """
    return np.string_(s)



def data_are_equal(d1, d2):
    """
    Simple utility to compare data in dicts
    
    Returns True only if all keys are the same, and all np.all data are the same
    """
    
    if set(d1) != set(d2):
        return False
    
    for k in d1:
        if not np.all(d1[k]==d2[k]):
            return False
        
    return True




#-----------------------------------------
# HDF5 utilities



def decode_attr(a):
    """
    Decodes:
        ASCII strings and arrays of them to str and arrays of str
        single-length arrays to scalar (Bmad writes this)
        
    """
    if isinstance(a, bytes):
        return a.decode('utf-8')
    
    if isinstance(a, np.ndarray):
        if a.dtype.type is np.string_:
            a = a.astype(str)
        if len(a) == 1:
            return a[0]
    
    return a  

def decode_attrs(attrs):
    return {k:decode_attr(v) for k,v in attrs.items()}


def encode_attr(a):
    """
    Encodes attribute
    
    See the inverse function:
        decode_attr
    
    """
    
    if isinstance(a, str):
        a = fstr(a)
    
    if isinstance(a, list) or isinstance(a, tuple):
        a = np.array(a)
    
    if isinstance(a, np.ndarray):
        if a.dtype.type is np.str_:
            a = a.astype(np.string_)
            
    return a

def encode_attrs(attrs):
    return {k:encode_attr(v) for k,v in attrs.items()}


import numpy as np
   
def write_bmad(particle_group,           
               outfile,
               p0c = None,
               t_ref = 0,
               verbose=False): 

    """
    Bmad's ASCII format is described in:
    
        https://www.classe.cornell.edu/bmad/manual.html
    
    Bmad normally uses s-based coordinates, with momenta:
        bmad px = px/p0
        bmad py = py/p0
        bmad pz = p/p0 - 1
    and longitudinal coordinate
        bmad z = -beta*c(t - t_ref)
    
    If p0c is given, this style of coordinates is written.
    
    Otherwise, Bmad's time based coordinates are written. 
    
    TODO: Spin
    
    """

    n = particle_group.n_particle            
    x = particle_group.x
    y = particle_group.y

    px = particle_group.px
    py = particle_group.py

    t = particle_group.t
    
    status  = particle_group.status 
    weight  = particle_group.weight
    
    zeros = np.full(n, 0)
    
    if p0c:
        # s-based coordinates
        
        # Check that z are all the same
        unique_z = np.unique(particle_group.z)
        assert len(unique_z) == 1, 'All particles must be a the same z position'        
        
        px = px/p0c
        py = py/p0c
        
        z = -particle_group.beta*299792458*(t - t_ref)
        pz = particle_group.p/p0c -1.0
        
        
    else:
        # Time coordinates.
        z = t
        pz = particle_group.pz  
        
 
    header = f"""!ASCII::3
0 ! ix_ele, not used
1 ! n_bunch
{n} ! n_particle
BEGIN_BUNCH
{particle_group.species} 
{particle_group.charge}  ! bunch_charge
0 ! z_center
0 ! t_center"""
    
    
    # <x> <px> <y> <py> <z> <pz> <macro_charge> <state> <spin_x> <spin_y> <spin_z>
    
    
    fmt = 7*['%20.12e'] + 1*['%2i']
    
    dat = np.array([x, px, y, py, z, pz, weight, status]).T
    
    footer = """END_BUNCH"""

    
    
    if verbose:
        print(f'writing {n} particles in Bmad ASCII format to {outfile}')
        
    # Write to file        
    np.savetxt(outfile, dat, header=header, footer=footer, fmt=fmt, comments='')
        



from pmd_beamphysics.units import c_light
import numpy as np

   
def write_litrack(particle_group,           
               outfile='litrack.zd',
               p0c=None,
               verbose=False): 

    """
    LiTrack is a Matlab code for the longitudinal phase space only.
    
   ASCII with columns:
   
   c*t (mm), relative energy deviation (%). 
   
   The head of the bunch has smaller ct
   
   
   bunch head is at smaller z.  the relative energy deviation unit is really %, which means I time 100 from relative energy spread.
    
    This routine makes ASCII particles, with column labels:
        'x', 'y', 'z', 'GBx', 'GBy', 'GBz', 't', 'q', 'nmacro'
    in SI units. 
    
    For now, only electrons are supported.

    """

    P = particle_group # convenience
    
    assert P.species == 'electron' # TODO: add more species
    
    assert np.all(P.weight > 0), 'ParticleGroup.weight must be > 0'
    
    n = particle_group.n_particle
    z = np.unique(P['z'])
    assert len(z) == 1, 'All particles must be a the same z. Please call .drift_to_z()'
    
    if p0c is None:
        p0c = P['mean_p']
        if verbose:
            print(f'Using mean_p as the reference momentum: {p0c} eV/c')
        
    
    ct = c_light*P.t
    delta = P.p/p0c -1.0 
    
    
    header = f"""% LiTrack particles
% 
% Created using the openPMD-beamphysics Python package
% https://github.com/ChristopherMayes/openPMD-beamphysics
%
% species: {P['species']}
% n_particle: {n}
% total charge: {P.charge} (C)
% reference momentum p0: {p0c} (eV/c)
%
% Columns: ct, delta = p/p0 -1
% Units: mm, percent"""

    outdat = np.array([ct*1000, delta*100]).T
    
    if verbose:
        print(f'writing {n} LiTrack particles to {outfile}')
    
    # Write ASCII
    np.savetxt(outfile, outdat, header=header, comments='', fmt = '%20.12e')
    
    return outfile
    
